/*****************************************************************************
 Copyright (c) 2004, Intel Corporation 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of Intel Corporation nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

****************************************************************************/

#include"softcore.h"
#include"hamcore.h"
#include"softserial.h"
#ifdef LINUX_DIAG
#include "afedebug.h"
#endif

extern struct global G;

int AllocateDebugResources (int dma_size);
//=============================================================================
int softserial_ioctl(struct tty_struct* ptty,
                      struct file* pfile,
                      unsigned int ioctl,
                      unsigned long ioctl_parameter)
{
  int error;
  unsigned int user_arg;
#ifdef LINUX_DIAG
  unsigned long iobase;
  unsigned long membase;
  unsigned long memlength;

  unsigned char  ch_value;
  unsigned short sh_value;
  unsigned int  l_value;

  AFEregister_t afereg;
  PCICfgData_t  PCIcfgReg;
  void *pBaseAddr;
#endif

  //int i;

  // printk("softserial:softserial_ioctl()\n");
  //   printk(
  //   "ham:ioctl[%X]: dir[%d], type[%d], nr[%d], size[%d]\n",
  //           ioctl,
  //           _IOC_DIR(ioctl),
  //           _IOC_TYPE(ioctl),
  //           _IOC_NR(ioctl),
  //           _IOC_SIZE(ioctl));
      
#ifdef LINUX_DIAG
   iobase = pci_resource_start(devAfe->pcidev, 1);
   membase = pci_resource_start(devAfe->pcidev, 0);
   memlength = pci_resource_len(devAfe->pcidev, 0);
   pBaseAddr = ioremap(membase, memlength); 
#endif

   switch (ioctl)
   {
#ifdef LINUX_DIAG
   case IOCTL_ALLOCATE_BUFFERS:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_ALLOCATE_BUFFERS\n");
      if(AllocateDebugResources(9800) < 0)
      {
          printk(KERN_INFO"softserial_ioctl: AllocateDebugResources FAIL\n"); 
          return(-1);
      }
      return(0);

   case IOCTL_FREE_BUFFERS:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_FREE_BUFFERS\n");
      if (michnld->bdbar != NULL)
      {
          pci_free_consistent(devAfe->pcidev, michnld->dma_size, michnld->bdbar,
                              michnld->bdbar_addr);
      }
      else
      {
          printk(KERN_INFO"softserial_ioctl: Free michnl Buffer FAIL\n"); 
          return(-1);
      }

      if (mochnld->bdbar != NULL)    
      {
          pci_free_consistent(devAfe->pcidev, mochnld->dma_size, mochnld->bdbar,
                              michnld->bdbar_addr);
      }
      else
      {
          printk(KERN_INFO"softserial_ioctl: Free mochnl Buffer FAIL\n"); 
          return(-1);
      }
      return(0);

   case IOCTL_GET_TXBUF_ADDR:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_GET_TXBUF_ADDR\n");

      { dmachnl_t *pres = (dmachnl_t*) ioctl_parameter;
        copy_to_user(pres, mochnld, sizeof(dmachnl_t));
      } 
      return(0);

   case IOCTL_GET_RXBUF_ADDR:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_GET_RXBUF_ADDR\n");

      { dmachnl_t *pres = (dmachnl_t*) ioctl_parameter;
        copy_to_user(pres, michnld, sizeof(dmachnl_t));
      } 
      return(0);

   case IOCTL_SET_TXBUF_DATA:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_SET_TXBUF_DATA\n");
      {  
         //unsigned int *pdata;
         copy_from_user(&dmabuf, (void*)ioctl_parameter, sizeof(dmabuffer_t));
         copy_from_user(mochnld->bdbar, dmabuf.buffer, 4*dmabuf.length);
         //pdata = (unsigned int*)mochnld->bdbar;
         //for (i=0;i<192;i++)
         //  printk(KERN_INFO"mochnld->bdbar[%d]=%X\n",i,pdata[i]); 
      }
      return(0);

   case IOCTL_GET_RXBUF_DATA:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_GET_RXBUF_DATA\n");
      {
         copy_from_user(&dmabuf, (void*)ioctl_parameter, sizeof(dmabuffer_t));
         copy_to_user(dmabuf.buffer, michnld->bdbar, 4*dmabuf.length);
         copy_to_user((void*)ioctl_parameter, &dmabuf, sizeof(dmabuffer_t));
      }
      return(0);

   case IOCTL_GET_AFE_8BIT_REG:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_GET_AFE_8BIT_REG\n");
    
      copy_from_user(&afereg, (void*)ioctl_parameter, sizeof(AFEregister_t));
      //save_flags(flags);
      //cli();
      afereg.value = (unsigned char)inb(iobase + afereg.address); 
      //restore_flags(flags);
      copy_to_user((void*)ioctl_parameter, &afereg, sizeof(AFEregister_t));
      return(0);
      
   case IOCTL_SET_AFE_8BIT_REG:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_SET_AFE_8BIT_REG\n");

      copy_from_user(&afereg, (void*)ioctl_parameter, sizeof(AFEregister_t)); 
      //save_flags(flags);
      //cli();
      outb((unsigned char)afereg.value, iobase + afereg.address); 
      //restore_flags(flags);
      return(0);

   case IOCTL_GET_AFE_32BIT_REG:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_GET_AFE_32BIT_REG\n");
    
      copy_from_user(&afereg, (void*)ioctl_parameter, sizeof(AFEregister_t));
      //save_flags(flags);
      //cli();
      afereg.value = inl(iobase + afereg.address); 
      //restore_flags(flags);
      copy_to_user((void*)ioctl_parameter, &afereg, sizeof(AFEregister_t));
      return(0);

   case IOCTL_SET_AFE_32BIT_REG:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_SET_AFE_32BIT_REG\n");

      copy_from_user(&afereg, (void*)ioctl_parameter, sizeof(AFEregister_t)); 
      //save_flags(flags);
      //cli();
      outl(afereg.value, iobase + afereg.address); 
      //restore_flags(flags);
      return(0);

   case IOCTL_GET_AFE_8BIT_MEM:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_GET_AFE_8BIT_MEM\n");

      copy_from_user(&afereg, (void*)ioctl_parameter, sizeof(AFEregister_t));
      //save_flags(flags);
      //cli();
      afereg.value = (unsigned char)readb(pBaseAddr + afereg.address);
      //restore_flags(flags);
      copy_to_user((void*)ioctl_parameter, &afereg, sizeof(AFEregister_t));
      iounmap(pBaseAddr); 
      return(0);

   case IOCTL_SET_AFE_8BIT_MEM:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_SET_AFE_8BIT_MEM\n");

      copy_from_user(&afereg, (void*)ioctl_parameter, sizeof(AFEregister_t));
      //save_flags(flags);
      //cli();
      writeb((unsigned char)afereg.value, pBaseAddr + afereg.address);
      //restore_flags(flags);
      iounmap(pBaseAddr); 
      return(0);

   case IOCTL_GET_AFE_32BIT_MEM:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_GET_AFE_32BIT_MEM\n");

      copy_from_user(&afereg, (void*)ioctl_parameter, sizeof(AFEregister_t));
      //save_flags(flags);
      //cli();
      afereg.value = readl(pBaseAddr + afereg.address);
      //restore_flags(flags);
      copy_to_user((void*)ioctl_parameter, &afereg, sizeof(AFEregister_t));
      iounmap(pBaseAddr); 
      return(0);

   case IOCTL_SET_AFE_32BIT_MEM:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_SET_AFE_32BIT_MEM\n");

      copy_from_user(&afereg, (void*)ioctl_parameter, sizeof(AFEregister_t));
      //save_flags(flags);
      //cli();
      writel(afereg.value, pBaseAddr + afereg.address);
      //restore_flags(flags);
      iounmap(pBaseAddr); 
      return(0);

   case IOCTL_GET_AFE_CONFIG_DATA:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_GET_CONFIG_DATA\n");

      copy_from_user(&PCIcfgReg, (void*)ioctl_parameter, sizeof(PCICfgData_t));
      //printk(KERN_INFO" GET_CONFIG_DATA: .type =%d \n", PCIcfgReg.datatype);
      switch(PCIcfgReg.datatype)
      {
          case 1:
             //printk(KERN_INFO" GET_CONFIG_DATA: BYTE \n");
             //save_flags(flags);
             //cli();
             pci_read_config_byte(devAfe->pcidev, PCIcfgReg.address, &ch_value);
             //restore_flags(flags);
             PCIcfgReg.value = ch_value;
             //printk(KERN_INFO"PCIcfgReg.value[1] = %X\n",PCIcfgReg.value);
             break;

          case 2:
             //printk(KERN_INFO" GET_CONFIG_DATA: WORD \n");
             //save_flags(flags);
             //cli();
             pci_read_config_word(devAfe->pcidev, PCIcfgReg.address, &sh_value);
             //restore_flags(flags);
             PCIcfgReg.value = sh_value;
             //printk(KERN_INFO"PCIcfgReg.value[1] = %X\n",PCIcfgReg.value);
             break;

          case 3:
             //printk(KERN_INFO" GET_CONFIG_DATA: DWORD \n");
             //save_flags(flags);
             //cli();
             pci_read_config_dword(devAfe->pcidev, PCIcfgReg.address, &l_value);
             //restore_flags(flags);
             PCIcfgReg.value = l_value;
             //printk(KERN_INFO"PCIcfgReg.value[1] = %lX\n",PCIcfgReg.value);
             break;
      }
      copy_to_user((void*)ioctl_parameter, &PCIcfgReg, sizeof(PCICfgData_t));
      return(0);

   case IOCTL_SET_AFE_CONFIG_DATA:
      //printk(KERN_INFO"softserial_ioctl: IOCTL_SET_CONFIG_DATA\n");

      copy_from_user(&PCIcfgReg, (void*)ioctl_parameter, sizeof(PCICfgData_t));

      switch(PCIcfgReg.datatype)
      {
          case 1:
             //printk(KERN_INFO" SET_CONFIG_DATA: BYTE \n");
             //save_flags(flags);
             //cli();
             pci_write_config_byte(devAfe->pcidev, PCIcfgReg.address, (unsigned char)PCIcfgReg.value);
             //restore_flags(flags);
             break;

          case 2:
             //printk(KERN_INFO" SET_CONFIG_DATA: WORD \n");
             //save_flags(flags);
             //cli();
             pci_write_config_word(devAfe->pcidev, PCIcfgReg.address, (unsigned short)PCIcfgReg.value);
             //restore_flags(flags);
             break;

          case 3:
             //printk(KERN_INFO" SET_CONFIG_DATA: DWORD \n");
             //save_flags(flags);
             //cli();
             pci_write_config_dword(devAfe->pcidev, PCIcfgReg.address, PCIcfgReg.value);
             //restore_flags(flags);
             break;
      }
      return(0);
#endif

//CASES that (may) need implementing
   case TIOCSERCONFIG:
   case TIOCSERGETLSR:
   case TIOCSERGSTRUCT:
   case TIOCGICOUNT:
   case TIOCGSERIAL: //TODO: give user serial structure
   case TIOCSSERIAL: //TODO: update our serial structure
      return (0);

//CASES NOT IN SERIAL.C BUT TTY
   case TCGETS: // tty calls get_termios
   case TCSETS: // tty calls set_termios
   case TCFLSH: // tty calls flush_buffer for various cases
   case TCSETSF: //tty calls set_termios TERMIOS_FLUSH
        return (-ENOIOCTLCMD);

//CASES FOR ONLY 2.1 KERNEL, IGNORE, RETURN AS DEFALUT
   case TCSBRK:
   case TCSBRKP:
   case TIOCGSOFTCAR:
   case TIOCSSOFTCAR:
      return (-ENOIOCTLCMD);

//OLD rejected DEBIAN call setserial -W     "setserial -W" is called in Debian boot
   case TIOCSERGWILD:
   case TIOCSERSWILD:
      return (0);

//CASES FOR MULTIPORT THAT DONT APPLY TO US
   case TIOCSERGETMULTI:
   case TIOCSERSETMULTI:
      return (-ENOIOCTLCMD);

//CASES implemented
   case TIOCMGET:
   {
      unsigned int status;
      status = 0;
   //   printk("softserial: TIOCMGET\n");
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0) 
      if(!MOD_IN_USE)
      {
         return -ENODEV;
      }
#else
      if(!G.refcount) return -ENODEV;
#endif
      return(put_user(G.softcore.get_line_status(0), (unsigned int*)ioctl_parameter));
   }

   case TIOCMBIS:
     // printk("softserial: TIOCMBIS\n");
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0) 
      if(!MOD_IN_USE) return -ENODEV;
#else
      if(!G.refcount) return -ENODEV;
#endif
      error = get_user(user_arg, (unsigned int*)ioctl_parameter);
      if (error) return (error);
      if (user_arg & TIOCM_RTS) G.softcore.rts_on();
      if (user_arg & TIOCM_DTR) G.softcore.dtr_on();
      return(0);

  case TIOCMBIC:
     // printk("softserial: TIOCMBIS\n");
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0) 
      if(!MOD_IN_USE) return -ENODEV;
#else
      if(!G.refcount) return -ENODEV;
#endif
      error = get_user(user_arg, (unsigned int*)ioctl_parameter);
      if (error) return (error);
      if (user_arg & TIOCM_RTS) G.softcore.rts_off();
      if (user_arg & TIOCM_DTR) G.softcore.dtr_off();
      return(0);


   case TIOCMSET:
      //printk("softserial: TIOCMSET\n");
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0) 
      if(!MOD_IN_USE)return -ENODEV;
#else
      if(!G.refcount) return -ENODEV;
#endif
      error = get_user(user_arg, (unsigned int*)ioctl_parameter);
      if (error) return (error);

      if (user_arg & TIOCM_RTS) G.softcore.rts_on();
      else                      G.softcore.rts_off();

      if (user_arg & TIOCM_DTR) G.softcore.dtr_on();
      else                      G.softcore.dtr_off();

      return(0);

   case TIOCMIWAIT: 
   {
      int ring_count, new_ring_count;
      int dsr_count, new_ndsr_count;
      int dcd_count, new_ndcd_count;
      int cts_count, new_ncts_count;
      //printk("softserial: TIOCMIWAIT\n");
     
      ring_count= G.softcore.get_line_status(1);//RINGING);
      dsr_count = G.softcore.get_line_status(2);//DSR);
      dcd_count = G.softcore.get_line_status(3);//RLSD);
      cts_count = G.softcore.get_line_status(4);//CTS);
      while (1)
      {
         /// wait until line status changes 
         G.softcore.msr_wait();
         
         /// if no signal pending, exit
         if (signal_pending(current))
            return (-ERESTARTSYS);
         
         new_ring_count = G.softcore.get_line_status(1);// RINGING);
         new_ndsr_count = G.softcore.get_line_status(2);// DSR);
         new_ndcd_count = G.softcore.get_line_status(3);// RLSD);
         new_ncts_count = G.softcore.get_line_status(4);// CTS);
         
         //* if no signals changed, return error 
         if ((new_ring_count == ring_count) && 
             (new_ndsr_count == dsr_count)  && 
             (new_ndcd_count == dcd_count)  && 
             (new_ncts_count == cts_count))
         {
            return (-EIO);
         }
         
      if (((ioctl_parameter & TIOCM_RNG) && (new_ring_count != ring_count))   ||
          ((ioctl_parameter & TIOCM_DSR) && (new_ndsr_count != dsr_count ))   ||
          ((ioctl_parameter & TIOCM_CD)  && (new_ndcd_count != dcd_count ))   ||
          ((ioctl_parameter & TIOCM_CTS) && (new_ncts_count != cts_count )))
         {
            return (0);
         }
      }
      break;
   }

   default:
      if(_IOC_TYPE(ioctl) == 't')  //some ppp ioctl command that we dont support
         return(-ENOIOCTLCMD);
      
     // printk("softserial: DEFAULT IOCTL\n");
      return (-ENOIOCTLCMD);
   }
   return (0);
}

//=============================================================================
#ifdef LINUX_DIAG
int AllocateDebugResources (int dma_size)
{
    michnld = &mi_chnld;
    mochnld = &mo_chnld;
    //printk(KERN_INFO"AllocateDebugResources: Entry\n");
    michnld->dma_size = dma_size;
    michnld->bdbar = pci_alloc_consistent(
                        devAfe->pcidev,
                        michnld->dma_size,
                        &michnld->bdbar_addr);
    if (michnld->bdbar == NULL) {
        return -ENOMEM;
    }
    memset(michnld->bdbar, 0, dma_size);

    mochnld->dma_size = dma_size;
    mochnld->bdbar = pci_alloc_consistent(
                        devAfe->pcidev,
                        mochnld->dma_size,
                        &mochnld->bdbar_addr);
    if (mochnld->bdbar == NULL) {
        return -ENOMEM;
    }
    memset(mochnld->bdbar, 0, dma_size);

    return 0;
}
#endif

